# local.py
#
# Copyright 2020-2024 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

import types
import os
import json
import hashlib
import sqlite3
import importlib.util
import datetime
import time
from gettext import gettext as _

from gi.repository import GLib

from .settings import Settings
from .fetcher import Fetcher
from . import remotes
from . import models


# # init and migration machinery

class UnitializedError(Exception):
    pass


class LocalDBDoesNotExistsException(Exception):
    pass


class LocalDBNotOpen(Exception):
    pass


dbfile: Optional[str] = None
db: Optional[sqlite3.Connection] = None
appdir: Optional[str] = None
cachedir: Optional[str] = None


def getDb():
    return db


def init(app_id: str):
    """Init paths and open db"""
    global appdir, cachedir

    appdir = os.path.join(GLib.get_user_data_dir(), app_id)
    if not os.path.exists(appdir):
        os.makedirs(appdir)

    cachedir = os.path.join(GLib.get_user_cache_dir(), app_id)
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)

    Settings.init(app_id)


def import_migration(filename: str) -> types.ModuleType:
    modulename = os.path.basename(filename).replace(".py", "")
    spec = importlib.util.spec_from_file_location(modulename, filename)
    if spec is None or spec.loader is None:
        raise ImportError(f"Unable to import migration {filename}")
    migration = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(migration)
    return migration


def migrate():
    schema = -1
    migrated = False
    try:
        c = db.execute("SELECT value FROM meta WHERE key='schema'")
        info = c.fetchone()
        if info is not None:
            schema = int(info['value'])
    except Exception as e:
        print("... ", e)

    migrationsdir = os.path.join(os.path.dirname(__file__), "migrations")
    files = sorted([f for f in os.listdir(migrationsdir) if f.endswith(".py")])

    try:
        c = db.cursor()
        for f in files:
            newschema = int(f.split("_")[0])
            if newschema > schema:
                # TODO: how to log using glib?
                print("Migrating to {}".format(f.replace(".py", "")))
                import_migration(os.path.join(migrationsdir, f)).up(c)
                schema = newschema
                migrated = True
    except Exception as e:
        db.rollback()
        raise e from None

    db.execute("INSERT OR REPLACE INTO meta (key, value) VALUES ('schema', ?)", (schema,))
    db.commit()

    return migrated

# event database


def openconf(conf: models.Conference, is_online=True):
    """Open event database for `conf`

    conf is a models.Conference
    After opening, tries to update meta
    """

    def _update_meta(*args):
        from .models import Meta
        with Meta() as m:
            m.title = m.title if m.title else conf.title
            m.start = m.start if m.start else conf.start
            m.end = m.end if m.end else conf.end

    f = opendb(conf.url, is_online)
    if f is None:
        _update_meta()
    else:
        f.connect("done", _update_meta)

    return f


def existsdb(url):
    """Check if exists database for `url`"""
    dbhashname = hashlib.md5(url.encode('utf8')).hexdigest()
    dbfilename = os.path.join(cachedir, dbhashname + ".db")
    return os.path.exists(dbfilename)


def opendb(url, is_online=True):
    """Open event database for `url`

    if `is_online`, fetch original pentabarf and update cache

    return Fetcher instance if downloading, else None
    """
    global dbfile, db
    dbhashname = hashlib.md5(url.encode('utf8')).hexdigest()
    dbfilename = os.path.join(cachedir, dbhashname + ".db")
    if not os.path.exists(dbfilename) and not is_online:
        db = None
        raise LocalDBDoesNotExistsException(_("Device disconnected and no cached data available."))

    print("open db", dbfilename)
    db = sqlite3.connect(dbfilename, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
    db.row_factory = sqlite3.Row

    migrated = migrate()

    dbfile = dbfilename

    from .models import Meta

    is_time_to_update = (time.time() - Meta().last_update) > Settings.instance().get_event_cache()

    if is_online and (migrated or is_time_to_update):
        return updatedb(url)

    return None


def updatedb(from_url):
    """Update event cache from schedule url

    return Fetcher instance
    """
    if db is None:
        raise LocalDBNotOpen()
    return remotes.update_schedule(from_url)


def close():
    """Close db"""
    global db, dbfile
    dbfile = None
    if db is not None:
        db.close()
        db = None


# # schedules list (from giggity and user defined)


def update_menu() -> Fetcher:
    """Fetch remote menu json"""
    # TODO: what about user entries shadowing entries from giggity?
    if cachedir is None:
        raise UnitializedError("cachedir is not defiend, local db not initialized")
    ggcachefile = os.path.join(cachedir, "ggmenu.json")
    url = "https://ggt.gaa.st/menu.json"
    return Fetcher(url, ggcachefile)


def open_menu(is_online):
    """
    Update menu cache file if expired or not exists.
    returns None or Fetcher
    """
    ggcachefile = os.path.join(cachedir, "ggmenu.json")
    print("menu cache file:", ggcachefile)

    need_refresh = not os.path.exists(ggcachefile)
    need_refresh = need_refresh or ((time.time() - os.path.getmtime(ggcachefile)) > Settings.instance().get_list_cache())
    if need_refresh and is_online:
        return update_menu()

    return None


def get_menu():
    """
    Get menu entries from menu cache and from user menu entries.
    User entries with same urls as entries from gg cache file shadow
    the latter (happens when user edit an entry from gg)
    """
    ggcachefile = os.path.join(cachedir, "ggmenu.json")
    userfile = os.path.join(cachedir, "usermenu.json")

    schedules = []
    if os.path.exists(userfile):
        with open(userfile, "r") as f:
            schedules += json.load(f)
    userurls = [s['url'] for s in schedules]

    if os.path.exists(ggcachefile):
        with open(ggcachefile, "r") as f:
            menu = json.load(f)
        for s in menu['schedules']:
            if s['url'] not in userurls:
                schedules.append(s)

    def _sort_key(elm):
        start = elm['start']
        return datetime.date(*[int(s) for s in start.split("-")])

    schedules = sorted(schedules, key=_sort_key, reverse=True)
    return schedules


def _add_user_menu_ready(sender, local_file_path, cbk):
    userfile = os.path.join(cachedir, "usermenu.json")
    from .models import Meta
    m = Meta()
    item = m.to_json()

    if m.title is None:
        m.title = ""
        m.save()

    schedules = []
    if os.path.exists(userfile):
        with open(userfile, "r") as f:
            schedules = json.load(f)

    schedules.append(item)
    with open(userfile, "w") as f:
        json.dump(schedules, f)

    cbk(m)
    close()


class MenuItemAlreadyExistsException(Exception):
    def __init__(self, message, obj):
        super().__init__(message)
        self.obj = obj


def add_user_menu(url, cbk):
    """
    Add and fetch schedule from url.
    Returns Fetcher or None.
    When ready calls "cbk({..metadata..})"
    """
    # try to not add same event twice:
    objs = [e for e in get_menu() if e['url'] == url]
    if len(objs) > 0:
        raise MenuItemAlreadyExistsException(_("Event already in list"), objs[0])

    f = opendb(url, True)  # I suppose that we are online at this point..
    if f is not None:
        f.connect("done", _add_user_menu_ready, cbk)
    else:
        _add_user_menu_ready(None, None, cbk)

    return f


def update_user_menu(data):
    """
    Update user menu entry
    data is a dict()
    """
    data['user'] = True
    userfile = os.path.join(cachedir, "usermenu.json")
    if os.path.exists(userfile):
        with open(userfile, "r") as f:
            schedules = json.load(f)
    else:
        schedules = []

    _id = None
    for i, s in enumerate(schedules):
        if s['url'] == data['url']:
            _id = i
            break
    if _id is None:
        schedules.append(data)
    else:
        schedules[_id] = data

    with open(userfile, "w") as f:
        json.dump(schedules, f)


def delete_user_menu(data):
    """
    Remove user menu entry
    `data` is a dict()
    """
    userfile = os.path.join(cachedir, "usermenu.json")
    if not os.path.exists(userfile):
        # non dovrei arrivare qui. solo i meta con user=True
        # posso essere editati, e quindi deve esistere il file
        raise Exception("User menu file not found. This should not happen.")

    with open(userfile, "r") as f:
        schedules = json.load(f)

    _id = None
    for i, s in enumerate(schedules):
        if s['url'] == data['url']:
            _id = i
            break
    if _id is not None:
        del schedules[_id]

    with open(userfile, "w") as f:
        json.dump(schedules, f)


# async image fetcher
def get_image_async(fileurl, cbk):
    """
    Save a remote image in local cache.
    Returns Fetcher or None
    When ready calls "cbk(None, 'localfilename')"
    """
    if fileurl is None:
        return None
    localfilename = hashlib.md5(fileurl.encode('utf8')).hexdigest()
    localdir = os.path.join(cachedir, "images")
    if not os.path.exists(localdir):
        os.makedirs(localdir)
    localfile = os.path.join(localdir, localfilename)
    if not os.path.exists(localfile):
        try:
            return Fetcher(fileurl, localfile, cbk)
        except Exception:
            return None
    cbk(None, localfile)
    return None


# cache handling
def get_cache_size():
    """get total size of cache files: db and images.
    skip json menu files"""
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(cachedir):
        for f in filenames:
            if not f.endswith(".json"):
                fp = os.path.join(dirpath, f)
                # skip if it is symbolic link
                if not os.path.islink(fp):
                    total_size += os.path.getsize(fp)
    return total_size


def clear_cache():
    """remove all files in cachedir except currently opened database and json menu files"""
    for dirpath, dirnames, filenames in os.walk(cachedir):
        for f in filenames:
            if not f.endswith(".json"):
                fp = os.path.join(dirpath, f)
                if fp != dbfile:
                    os.remove(fp)



import re
from html.parser import HTMLParser

from gi.repository import GLib


def convert_markdown_links(text):
    """converts markdown links to html hyperlinks"""
    return re.sub(r"\[([^]]+?)\]\(([^)\s]+?)(?:\s.*?)?\)", r'<a href="\2">\1</a>', text)


def clean_markup(text):
    """remove unsupported markup from text to use in label"""
    if text is None:
        text = ""

    text = html2pango(text)
    text = text.replace("\n\n\n", "\n\n")
    text = text.replace("&", "&amp;")

    return text.strip()


def markup_to_text(text):
    """return plain text version of markup string"""
    text = clean_markup(text)
    text = re.sub(r'</?i>', '_', text)
    text = re.sub(r'</?tt>', '`', text)
    text = re.sub(r'</?b>', '**', text)
    text = text.replace("&amp;", "&")

    # remove all remainig tags
    text = re.sub(r'</?\w+[^>]*>', '', text)
    return text


class InterruptibleTimeout:
    def __init__(self, intervall=0, callback=None, data=None):
        self._running = False
        self._requested_stop = True
        self._callback = callback
        if callback:
            self._running = True
            self._requested_stop = False
            if callback(self, data):
                GLib.timeout_add(intervall, self._on_timeout, data)

    def _on_timeout(self, data):
        if self._requested_stop:
            res = False
        else:
            res = self._callback(self, data)
        self._running = res
        return res

    def stop(self):
        self._requested_stop = True

    def must_stop(self):
        return self._requested_stop


# pango tags:
TAG_MAP = {
    # pango tags, as is
    'b': ('<b>', '</b>'),
    'i': ('<i>', '</i>'),
    's': ('<s>', '</s>'),
    'sub': ('<sub>', '</sub>'),
    'sup': ('<sup>', '</sup>'),
    'big': ('<big>', '</big>'),
    'small': ('<small>', '</small>'),
    'tt': ('<tt>', '</tt>'),
    'u': ('<u>', '</u>'),

    # blocks
    'address': ('', '\n\n'),
    'blockquote': ('', '\n\n'),
    'div': ('', '\n\n'),
    'footer': ('', '\n\n'),
    'header': ('', '\n\n'),
    # 'hr': ('', '<span qualcosa>'),
    'p': ('', '\n\n'),
    'pre': ('\n<tt>', '</tt>\n'),


    'h1': ('<big><b>', '</b></big>\n\n'),
    'h2': ('<b>', '</b>\n\n'),
    'h3': ('<b>', '</b>\n\n'),
    'h4': ('<small><b>', '</b></small>\n\n'),
    'h5': ('<small><b>', '</b></small>\n\n'),
    'h6': ('<small><b>', '</b></small>\n\n'),
    'td': ('<tt>', '</tt>'),
    'th': ('<b><tt>', '</tt></b>'),
    'tr': ('', '\n'),
    'dt': ('<b>', '<b>'),
    'dd': ('    ', '\n'),
    'li': ('', '\n'),

    # inline
    'em': ('<i>', '</i>'),
    'var': ('<i><tt>', '</tt></i>'),
    'kbd': ('<i><tt>', '</tt></i>'),
    'code': ('<tt>', '</tt>'),
    'strong': ('<b>', '</b>'),
    # a
    'ins': ('<u>', '</u>'),
    'del': ('<s>', '</s>'),
    'strike': ('<s>', '</s>'),
    'br': ('', '\n'),
}


class HTMLToPango(HTMLParser):
    def __init__(self):
        super().__init__()
        self.result = []

    def a_s(self, tag, attrs):
        href = attrs.get('href', '')
        self.result.append(f'<a href="{href}">')

    def a_e(self, tag):
        self.result.append('</a>')

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag in TAG_MAP:
            self.result.append(TAG_MAP[tag][0])
            return

        f = getattr(self, f"{tag}_s", None)
        if f:
            f(tag, attrs)

    def handle_endtag(self, tag):
        if tag in TAG_MAP:
            self.result.append(TAG_MAP[tag][1])
            return

        f = getattr(self, f"{tag}_e", None)
        if f:
            f(tag)

    def handle_data(self, data):
        self.result.append(data.strip())


def html2pango(text):
    parser = HTMLToPango()
    parser.feed(text)
    return ''.join(parser.result)



# widgets/loaderpage.py
#
# Copyright 2020-2023 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Adw

from ..fetcher import Fetcher


@Gtk.Template(resource_path="/net/kirgroup/confy/widgets/loaderpage.ui")
class LoaderPage(Adw.Bin):
    __gtype_name__ = "ConfyLoaderPage"

    progressbar = Gtk.Template.Child()  # type: ignore
    cancel_button = Gtk.Template.Child()  # type: ignore

    fetcher: Optional[Fetcher] = None

    def __init__(self):
        super().__init__()
        GLib.timeout_add(200, self._do_pulse)

    @Gtk.Template.Callback()
    def on_cancel_button_clicked(self, *_):
        if self.fetcher:
            self.fetcher.cancel()

    def set_fetcher(self, fetcher):
        self.fetcher = fetcher
        self.progressbar.set_fraction(0)
        fetcher.bind_property("fraction", self.progressbar, "fraction")

    def _do_pulse(self, *args):
        if self.progressbar.get_fraction() > 0:
            return False
        self.progressbar.pulse()
        return True



# mainwindow.py
#
# Copyright 2020-2024 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

import json
from time import time
from gettext import gettext as _

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Adw
from gi.repository import Notify

from . import local
from . import models
from .settings import Settings
from .openwindow import OpenWindow
from .config import APP_NAME, APP_ID  # type:ignore # config is generated at compile time
from .networkconnectionmonitor import NetworkConnectionMonitor
from .widgets import LoaderPage
from .widgets import Navbar
from .widgets.navbar import NavbarItem
from .pages import PageInfo
from .pages import PageList
from .pages import PageEvents
from .utils import clean_markup

PAGES = dict(
    info=NavbarItem(name="info", label=_("Event info"), icon="emblem-system-symbolic"),
    days=NavbarItem(name="days", label=_("Days"), icon="x-office-calendar-symbolic"),
    tracks=NavbarItem(name="tracks", label=_("Tracks"), icon="format-justify-fill-symbolic"),
    rooms=NavbarItem(name="rooms", label=_("Rooms"), icon="display-projector-symbolic"),
    starred=NavbarItem(name="starred", label=_("Starred"), icon="starred-symbolic", group=1)
)


@Gtk.Template(resource_path="/net/kirgroup/confy/mainwindow.ui")
class MainWindow(Adw.ApplicationWindow):
    __gtype_name__ = "ConfyMainWindow"

    @GObject.Signal
    def conf_updated(self):
        pass

    @GObject.Property(type=bool, default=False)
    def has_recent(self):
        return self.recent_model.get_n_items() > 0

    overlay: Adw.ToastOverlay = Gtk.Template.Child()  # type: ignore
    # stack to switch between 'start', 'loader' and 'main' views
    loader_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    # the loader widget inside 'loader' view
    loader: LoaderPage = Gtk.Template.Child()  # type: ignore

    # main split view
    main_split_view: Adw.OverlaySplitView = Gtk.Template.Child()  # type: ignore

    # sidebar navbar
    navbar: Navbar = Gtk.Template.Child()  # type: ignore

    # main navigation view
    main_stack: Adw.NavigationView = Gtk.Template.Child()  # type: ignore

    # recent confs model and menu, in 'start' page
    recent_model: Gtk.NoSelection = Gtk.Template.Child()  # type: ignore
    recent_menu: Gio.MenuItem = Gtk.Template.Child()  # type: ignore

    # Monitor network connectivity
    nm = NetworkConnectionMonitor()

    # current opened conference
    conf: Optional[models.Conference] = None

    # dict of sent notifications.
    # Keeps Notification ojects around to be able to get activation events
    notifications: dict[str, Notify.Notification] = {}

    def __init__(self, application: Gtk.Application, **kwargs):
        super().__init__(application=application, **kwargs)

        size = Settings.instance().get_size()
        self.set_default_size(*size)

        # Check if there are notifications to send
        application.connect('tick', self.check_event_notification)

        # handle conf data updating, eg. reloading from remote
        self.connect('conf-updated', self._on_conf_updated)

        self._add_action("open-conference", None, ["<Primary>o",], self.open_conference)
        self._add_action("search", None, ["<Primary>f",], lambda *_: self.show_search_all())
        self._add_action("copy", None, ["<Primary>c",], lambda *_: self.copy_details_to_clipboard())
        # open recent action, used by recent menu
        self._add_action("open-recent", "i", None, self._on_recent_action)

        # TODO: serve?
        self.nm.connect("notify", self._on_networkconnectionmonitor_notify)

        model = Gio.ListStore.new(models.RecentConf)
        self.recent_model.set_model(model)

        self.recentmanager = Gtk.RecentManager.get_default()
        self.recentmanager.connect("changed", self._on_recent_changed)
        self._on_recent_changed()

    def _add_action(self, name, parameter_type, accels, activate_cbk):
        """Simplify adding an action to the window"""
        if parameter_type is not None:
            parameter_type = GLib.VariantType.new(parameter_type)
        action = Gio.SimpleAction.new(name, parameter_type)
        action.connect("activate", activate_cbk)
        self.add_action(action)
        if accels is not None:
            self.get_application().set_accels_for_action(F"win.{name}", accels)

        return action

    def _on_networkconnectionmonitor_notify(self, *_):
        # TODO: what should be here?
        ...

    def _on_recent_changed(self, *args):
        # empty recent confs model and menu...
        model = self.recent_model.get_model()
        menu = self.recent_menu
        menu.remove_all()
        while model.get_n_items() > 0:
            model.remove(0)

        # ... then rebuild model and menu
        n = 0
        for item in self.recentmanager.get_items():
            if item.has_application(APP_NAME):
                print(f"#{n} {item.get_uri()} {item.get_applications()}")
                recentconf = models.RecentConf(item)
                model.append(recentconf)
                # add menu item, trigger 'open-recent' action
                menu.append(recentconf.props.title, f"win.open-recent({n})")
                n = n + 1
        self.notify("has-recent")

    @Gtk.Template.Callback()
    def _on_recent_activated(self, listview, index):
        """Called by recent listview 'activated' signal"""
        recentconf = self.recent_model.get_item(index)
        conf = json.loads(recentconf.data)
        conf = models.Conference()._init(conf)
        # don't load conf if it is already loaded
        if self.conf is not None and self.conf.url == conf.url:
            print(f"Same url {self.conf.url} == {conf.url}. Skip")
            return
        self.load_conf(conf)

    def _on_recent_action(self, action, args):
        """
        Called by open-recent action activation

        first arg from action is int32 index in recent model
        """
        index = args.get_int32()
        self._on_recent_activated(None, index)

    def _add_recent(self, conf: models.Conference):
        """
        Add conference to recent.
        """

        # remove current conf if is already in recents
        # this way will be readded on top
        # TODO: is there a better way to do this?
        if self.recentmanager.has_item(conf.url):
            self.recentmanager.remove_item(conf.url)

        recentdata = Gtk.RecentData()
        recentdata.mime_type = "x-application/confy-event-uri"
        recentdata.app_exec = "confy"
        recentdata.is_private = True
        recentdata.app_name = APP_NAME
        recentdata.display_name = conf.title
        # misuse recentdata description to hold conf object, yay!
        recentdata.description = json.dumps(conf.to_json())

        self.recentmanager.add_full(conf.url, recentdata)

    def _send_desktop_notification(self, title: str, body: str, nid: str):
        """
        Send a notification and register it to react to user action
        """
        n = Notify.Notification.new(title, body, APP_ID)
        n.set_timeout(Notify.EXPIRES_NEVER)
        n.add_action("view", "View", self._on_notification_action, nid)
        n.show()
        self.notifications[nid] = n

    def _on_conf_updated(self, *args):
        """
        Callback when conference data is modified (eg. is reloaded)
        """

        self.navbar.set_item_cond(PAGES['info'], True)
        self.navbar.set_item_cond(PAGES['days'], models.Day.count() > 0)
        self.navbar.set_item_cond(PAGES['tracks'], models.Track.count() > 0)
        self.navbar.set_item_cond(PAGES['rooms'], models.Room.count() > 0)
        self.navbar.set_item_cond(PAGES['starred'], True)

    def _on_fetcher_done(self, *_):
        """
        Callback when conference has been open from remote url
        """
        self._add_recent(self.conf)
        self.emit("conf_updated")
        self.show_main()

    def _on_fetcher_error(self, fetcher, error):
        """
        Callback when there was an error fetching conference from remote url
        """
        self.show_error(error)

    def _on_notification_action(self, notification, action, data):
        """
        User clicked on "View" in notification
        """
        print(notification, action, data)
        if action == "view" and data.startswith("nextup-"):
            self.present_with_time(time())
            eventid = int(data.split("-")[1])
            event = models.Event.by_id(eventid)
            # if we are not on "events" subpage, force it as non navigatable,
            # we want to be back to where we were
            print(self.subpage_leaflet.get_visible_child(), self.subpage_events)
            can_we_go_back_to_events = self.subpage_leaflet.get_visible_child() == self.subpage_events
            self.show_details(event, can_we_go_back_to_events)

    @Gtk.Template.Callback()
    def _on_page_changed(self, navbar):
        self.navigate_to(navbar.get_current_item().name)

    @Gtk.Template.Callback()
    def _on_close_requested(self, *args):
        print("close-request", args)

    def open_conference(self, *_):
        """
        Show open window to open a conference
        """
        app = self.get_application()
        win = OpenWindow(modal=True, transient_for=self, application=app)
        win.show()

    def load_conf(self, conf: models.Conference):
        """
        Load conference

        Setup/reset window to display `conf`
        """
        self.conf = conf
        # we reset navigation because 'conf-updated' is emited, but
        # we don't really have a clear way to get the updated conf in pages
        # as we cannot attach a conf to the signal, as is not a GObject.
        # So existing pages will update values but with old conf object.
        # Reset navigation will force to create a new PageInfo with the new
        # conf object
        self.navigate_to(None)  # reset navigation

        self.show_loader()
        try:
            fetcher = local.openconf(conf, is_online=self.nm.get_isconnected())
        except Exception as e:
            self.show_error(e)
            return
        else:
            if fetcher is not None:
                # 'local.openconf' returns a `Fetcher` if there is no local cache
                # or the local cache is expired, and data must be downloaded
                # we listen to fetcher events
                fetcher.connect("done", self._on_fetcher_done)
                fetcher.connect("error", self._on_fetcher_error)
                # pass the fetcher to the loader to show progress
                self.loader.set_fetcher(fetcher)
            else:
                # there was a valid local cache, signal change in conf,
                # and show main page
                self._add_recent(self.conf)
                self.emit("conf-updated")
                self.show_main()

        self.notifications = {}
        self.check_event_notification()

    def check_event_notification(self, *args):
        """
        Notifiy starred talk staring in next 5 minutes
        """
        if self.conf is None:
            # no conf loaded, nothing to do
            return
        nextup = list(models.Event.nextup(5))
        for e in nextup:
            title = clean_markup(e.title)
            self._send_desktop_notification(
                _("Next up: {}").format(title),
                _("at {} in {}").format(
                    e.start_in_tz().strftime("%H:%M"),
                    e.room
                ),
                f"nextup-{e.id}")
            e.set_notified(True)

    def show_error(self, error=None):
        msg = Adw.MessageDialog(
            transient_for=self,
            modal=True,
            heading=_("Error loading event"),
            body=str(error) if error else "",
        )
        msg.add_response("ok", _("Ok"))
        msg.connect("response", self.open_conference)
        msg.present()

    def show_start(self):
        """Show 'start' page"""
        self.loader_stack.set_visible_child_name("start")

    def show_loader(self):
        """Show 'loader' page"""
        self.loader_stack.set_visible_child_name("loader")

    def show_main(self):
        """Show 'main' page"""
        self.navigate_to("info")
        self.loader_stack.set_visible_child_name("main")

    def navigate_to(self, name: Optional[str]):
        print("Navigate to", name)
        if name is None:
            self.main_stack.replace([Adw.NavigationPage()])
            return

        # we already into this navigation path?
        top = self.main_stack.find_page(name)
        if top is not None:
            self.main_stack.replace_with_tags([name])
            if self.main_split_view.get_collapsed():
                self.main_split_view.set_show_sidebar(False)
            return

        # no, start anew
        child: Adw.NavigationPage
        match name:
            case "info":
                info = PAGES[name]
                child = PageInfo(self, self.conf, title=info.label, tag=name)
            case "starred" | "search":
                child = PageEvents(None, model=name, tag=name)
            case _:
                info = PAGES[name]
                child = PageList(self, self.conf, model=name, title=info.label, tag=name)

        self.main_stack.replace([child])
        if self.main_split_view.get_collapsed():
            self.main_split_view.set_show_sidebar(False)

    def copy_details_to_clipboard(self):
        page = self.main_stack.get_visible_page()
        if page and page.get_tag() == "details":
            page.copy_to_clipboard()
            self.notification_show(_("Event details copied to the clipboard"))

    def notification_show(self, text):
        t = Adw.Toast()
        t.set_title(text)
        self.overlay.add_toast(t)

    def show_search_all(self):
        self.navigate_to("search")



# widgets/navbar.py
#
# Copyright 2020-2024 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass

from gi.repository import GObject
from gi.repository import Gtk


class NavbarListItem(Gtk.ListBoxRow):
    __gtype_name__ = "ConfyNavbarListItem"

    def __init__(self, item, **kwargs):
        super().__init__(**kwargs)
        self.item = item

        box = Gtk.Box(spacing=8)
        icon = Gtk.Image(icon_name=item.icon)
        label = Gtk.Label(halign=Gtk.Align.START, label=item.label)
        box.append(icon)
        box.append(label)
        self.set_child(box)


@dataclass
class NavbarItem:
    name: str = ""
    icon: str = ""
    label: str = ""
    group: int = 0


class Navbar(Gtk.ListBox):
    __gtype_name__ = "ConfyNavbar"

    model: list[NavbarItem] = []
    _current_item = None  # type: ignore

    @GObject.Signal
    def page_changed(self):
        ...

    def __init__(self, **kwargs):
        super().__init__(hexpand=True)
        self.add_css_class("navigation-sidebar")
        self.connect("row-activated", self._on_nav)

    def update(self):
        self.remove_all()
        found = False
        for idx, item in enumerate(self.model):
            widget = NavbarListItem(item)
            if self._current_item and item == self._current_item:
                found = True
                self.select_row(widget)
            self.append(widget)

        if not found:
            self._current_item = self.model[0]
            self.select_row(self.get_row_at_index(0))
            self.emit("page-changed")

    def _on_nav(self, sender, row):
        self._current_item = row.item
        self.emit("page-changed")

    def get_current_item(self) -> NavbarItem:
        return self._current_item  # type: ignore

    def set_item_cond(self, item: NavbarItem, cond: bool):
        """
            if cond is true and item not in model, item is appended
            if cond is false and item is in model, item is removed
        """
        if cond:
            if item not in self.model:
                self.model.append(item)
                self.update()
        else:
            if item in self.model:
                self.model.remove(item)
                self.update()



# events.py
#
# Copyright 2020-2024 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Adw

from ..utils import convert_markdown_links, clean_markup, markup_to_text
from ..widgets import EventActionRow, Author, StarButton
from ..widgets import WebLinkActionRow, weblinkactionrowactivated
from ..models import Event


@Gtk.Template(resource_path="/net/kirgroup/confy/pages/details.ui")
class PageDetails(Adw.NavigationPage):
    __gtype_name__ = "ConfyPageDetails"

    scroll: Gtk.ScrolledWindow = Gtk.Template.Child()  # type: ignore
    star_button: StarButton = Gtk.Template.Child()  # type: ignore
    conflicts_list: Gtk.ListBox = Gtk.Template.Child()  # type: ignore
    links_list: Gtk.ListBox = Gtk.Template.Child()  # type: ignore
    author_box: Gtk.FlowBox = Gtk.Template.Child()  # type: ignore

    _subtitle = ""
    _authors = ""
    _time = ""
    _room = ""
    _track = ""
    _abstract = ""
    _description = ""

    @GObject.Property(type=str, default="")
    def subtitle(self):
        return self._subtitle

    @subtitle.setter
    def subtitle_setter(self, v):
        self._subtitle = v
        self.notify("has-subtitle")

    @GObject.Property(type=bool, default=False)
    def has_subtitle(self):
        return self._subtitle != ""

    @GObject.Property(type=str, default="")
    def authors(self):
        return self._authors

    @authors.setter
    def authors_setter(self, v):
        self._authors = v
        self.notify("has-authors")

    @GObject.Property(type=bool, default=False)
    def has_authors(self):
        return self._authors != ""

    @GObject.Property(type=str, default="")
    def time(self):
        return self._time

    @time.setter
    def time_setter(self, v):
        self._time = v

    @GObject.Property(type=str, default="")
    def room(self):
        return self._room

    @room.setter
    def room_setter(self, v):
        self._room = v

    @GObject.Property(type=str, default="")
    def track(self):
        return self._track

    @track.setter
    def track_setter(self, v):
        self._track = v

    @GObject.Property(type=str, default="")
    def abstract(self):
        return self._abstract

    @abstract.setter
    def abstract_setter(self, v):
        self._abstract = v
        self.notify("has-abstract")

    @GObject.Property(type=bool, default=False)
    def has_abstract(self):
        return self._abstract is not None and self._abstract != "" and self._abstract != "None"

    @GObject.Property(type=str, default="")
    def description(self):
        return self._description

    @description.setter
    def description_setter(self, v):
        self._description = v
        self.notify("has-description")

    @GObject.Property(type=bool, default=False)
    def has_description(self):
        return self._description is not None and self._description != "" and self._description != "None"

    _handler_id: Optional[int] = None
    obj: Optional[Event] = None

    def __init__(self, obj, **kwargs):
        super().__init__(**kwargs)
        self.set_obj(obj)

    def set_obj(self, obj):
        if self._handler_id:
            self.obj.disconnect(self._handler_id)
        self.obj = obj
        self._handler_id = obj.connect('update', self._on_obj_update)

        self.scroll.get_vadjustment().set_value(0)
        self._on_obj_update()

    def _on_obj_update(self, *_):
        self.props.title = markup_to_text(self.obj.title)
        self.props.authors = ", ".join([clean_markup(str(p)) for p in self.obj.persons()])
        self.props.subtitle = markup_to_text(self.obj.subtitle)
        self.props.time = "{}-{}".format(
            self.obj.start_in_tz().strftime("%A %H:%M"),
            self.obj.end_in_tz().strftime("%H:%M"))
        self.props.room = clean_markup(self.obj.room)
        self.props.track = clean_markup(self.obj.track)

        self.props.abstract = clean_markup(convert_markdown_links(self.obj.abstract))
        self.props.description = clean_markup(convert_markdown_links(self.obj.description))

        self.author_box.remove_all()
        for author in self.obj.persons():
            widget = Author(author)
            wrap = Gtk.FlowBoxChild(child=widget)
            self.author_box.append(wrap)

        self._conflicts = conflicts = list(self.obj.get_conflicts())
        self.conflicts_list.props.visible = len(conflicts) > 0
        w = self.conflicts_list.get_last_child()
        while w:
            self.conflicts_list.remove(w)
            w = self.conflicts_list.get_last_child()
        for c in conflicts:
            row = EventActionRow(c)
            self.conflicts_list.append(row)

        links = list(self.obj.links())
        self.links_list.props.visible = len(links) > 0
        w = self.links_list.get_last_child()
        while w:
            self.links_list.remove(w)
            w = self.links_list.get_last_child()
        for link in links:
            row = WebLinkActionRow(link.name, link.href)
            self.links_list.append(row)

        self.star_button.props.is_starred = self.obj.starred

    def copy_to_clipboard(self):
        cb = Gdk.Display.get_default().get_clipboard()
        data = [self.get_title()]

        def _append(txt):
            if txt != "":
                data.append(txt)

        _append(self._subtitle)
        _append(self._authors)
        data.append("")
        _append(self._time)
        _append(self._room)
        _append(self._track)
        data.append("")
        _append(self._abstract)
        _append(self._description)
        data.append("")
        for link in self.obj.links():
            data.append(f"{link.name}\n{link.href}\n")

        text = "\n".join(data)
        text = markup_to_text(text)
        print(text)
        cb.set(text)

    @Gtk.Template.Callback()
    def toggle_starred(self, *_):
        self.obj.set_star(not self.obj.starred)

    @Gtk.Template.Callback()
    def on_conflict_activated(self, listbox, row):
        self.set_obj(row.obj)

    @Gtk.Template.Callback()
    def on_link_activated(self, listbox, row):
        weblinkactionrowactivated(listbox, row)



# widgets/__init__.py
#
# Copyright 2020-2023 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .eventeditdialog import EventEditDialog
from .loaderpage import LoaderPage
from .searchbaroverlay import SearchBarOverlay
from .navbar import Navbar
from .starbutton import StarButton
from .datepickerrow import DatePickerRow
from .inputdialog import InputDialog
from .loadingoverlay import LoadingOverlay
from .eventactionrow import EventActionRow
from .weblinkactionrow import WebLinkActionRow, weblinkactionrowactivated
from .author import Author
from .togglesidebarbutton import ToggleSidebarButton



def up(cursor):
    cursor.execute("ALTER TABLE persons ADD COLUMN organization TEXT DEFAULT '';")
    cursor.execute("ALTER TABLE persons ADD COLUMN thumbnail TEXT DEFAULT '';")
    cursor.execute("ALTER TABLE persons ADD COLUMN bio TEXT DEFAULT '';")
    cursor.execute("ALTER TABLE persons ADD COLUMN url TEXT DEFAULT '';")


def down(cursor):
    ...



# widgets/eventactionrow.py
#
# Copyright 2020-2023 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional
import datetime
import math

from dateutil.tz import UTC

from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Adw
from gi.repository import Graphene

from ..utils import clean_markup
from ..models import dbUpdateWatcher, Event

from .starbutton import StarButton


class EventActionRow(Adw.ActionRow):
    _hid: Optional[int] = None
    obj: Optional[Event] = None

    def __init__(self, obj):
        super().__init__(activatable=True, selectable=False)
        self.obj = obj

        self._hid = dbUpdateWatcher.connect('update', self.on_update)

        self.toggle_btn = StarButton(is_starred=obj.starred)
        self.toggle_btn.connect("clicked", self.toggle_starred)
        self.add_suffix(self.toggle_btn)

        self.on_update()

        Gio.Application.get_default().connect(
            "tick", lambda *_: self.queue_draw()
        )

    def __del__(self):
        if self._hid:
            dbUpdateWatcher.disconnect(self.hid)

    def _is_first(self):
        return self.get_index() == 0

    def _is_last(self):
        p = self.get_parent()
        if p is None:
            return True
        return p.get_last_child() == self

    def do_snapshot(self, snapshot: Gtk.Snapshot):
        x = -2
        y = -2
        w = 10
        h = self.get_allocated_height()

        rect = Graphene.Rect().init(x, y, w, h)
        cr = snapshot.append_cairo(rect)
        self._draw(cr, rect)
        first_child = self.get_first_child()
        if first_child is not None:
            self.snapshot_child(first_child, snapshot)

    def _roundedclippath(self, cr, x, y, w, h, rtop, rbottom):
        cr.arc(x + rtop, y + rtop, rtop, math.pi, 3 * math.pi / 2)
        cr.line_to(x + w, y)
        cr.line_to(x + w, y + h)
        cr.line_to(x + w, y + h)
        # cr.arc(x+w-rtop, y+rtop, rtop, 3*math.pi/2, 0) # top-right round corner
        # cr.arc(x+w-rbottom, y+h-rbottom, rbottom, 0, math.pi/2) # bottom-right
        cr.arc(x + rbottom, y + h - rbottom, rbottom, math.pi / 2, math.pi)
        cr.close_path()
        cr.clip()

    def _draw(self, cr, rect):
        """
        Draw event progress indicator in widget background.
        It's a tick line on left border from top whose lenght is relative to the
        percentage of event time elapsed.

        We need to handle list round corners. Gtk doesn't clip element like HTML
        when rendering with "border-radius" value, so our line will overflow the
        corner.

        I use the clip path method. It get a top rounded corner if `self` is the first
        row in the list, and a bottom rounded corner if it's the last.
        Round radio is defined in code, as I can't find a way to get current widget CSS
        values.
        Then a line of required lenght, color and width is drawn from top.
        Color of the line is widget foreground color with alpha 0.2. Using foreground color
        make the line work on light and dark theme.
        """
        now = datetime.datetime.now(UTC).replace(tzinfo=None)
        # now = datetime.datetime(2023, 2, 5, 10, 50, tzinfo=UTC).replace(tzinfo=None)
        obj = self.obj

        dur = (obj.end.timestamp() - obj.start.timestamp())
        pos = (now.timestamp() - obj.start.timestamp())
        if dur > 0:
            prc = min(max(pos / dur, 0.0), 1.0)
        else:
            # not a valid event duration
            # set prc to 0 if the event is in the future or 1 if in the past
            prc = int(pos >= 0)

        if prc == 0:
            return

        maxh = rect.get_height()
        x = rect.get_x()
        y = rect.get_y()
        h = maxh * prc
        w = rect.get_width()

        # clip at rounded corners
        rtop = 12 if self._is_first() and self.get_header() is None else 0
        rbottom = 12 if self._is_last() else 0

        # color
        # >= gtk 4.10 : self.get_color()
        color = self.get_style_context().get_color()
        color.alpha = 0.2

        cr.save()

        # rounded clip path
        self._roundedclippath(cr, x, y - 1, w, maxh + 2, rtop, rbottom)

        # draw line
        cr.set_source_rgba(
            color.red, color.green, color.blue, color.alpha
        )
        cr.set_line_width(w)
        cr.move_to(x, y)
        cr.line_to(x, h)
        cr.stroke()
        cr.restore()

    def on_update(self, *args):
        obj = self.obj

        subtitle = "{}-{}".format(
            obj.start_in_tz().strftime("%a %H:%M"),
            obj.end_in_tz().strftime("%H:%M"))

        if obj.room:
            subtitle = "{} ({})".format(
                subtitle,
                obj.room)

        if obj.track:
            subtitle = "{} '{}'".format(
                subtitle,
                obj.track)

        ps = list(obj.persons())
        if len(ps) > 0:
            subtitle = "{} - {}".format(
                subtitle,
                ", ".join([str(p) for p in ps]))

        self.set_title(clean_markup(obj.title))
        self.set_subtitle(clean_markup(subtitle))

        self.toggle_btn.props.is_starred = self.obj.starred

    def toggle_starred(self, button):
        self.obj.set_star(not self.obj.starred)



def up(cursor):
    cursor.execute("""CREATE TABLE IF NOT EXISTS meta (
        key TEXT(20) PRIMARY KEY,
        value TEXT)""")
    cursor.execute("""CREATE TABLE IF NOT EXISTS persons (
        id TEXT(50) PRIMARY KEY,
        name TEXT)""")
    cursor.execute("""CREATE TABLE IF NOT EXISTS event_person (
        event_id TEXT(50),
        person_id TEXT(50),
        FOREIGN KEY(event_id) REFERENCES events(id),
        FOREIGN KEY(person_id) REFERENCES persons(id))""")
    cursor.execute("""CREATE TABLE IF NOT EXISTS events (
        id TEXT(50) PRIMARY KEY,
        date DATE,
        start TIMESTAMP,
        end TIMESTAMP,
        room TEXT DEFAULT "",
        slug TEXT,
        title TEXT,
        subtitle TEXT,
        track TEXT DEFAULT "",
        type TEXT,
        abstract TEXT,
        description TEXT,
        starred INTEGER(1) DEFAULT 0,
        notified INTEGER(1) DEFAULT 0)""")
    cursor.execute("CREATE INDEX IF NOT EXISTS event_date ON events(date)")
    cursor.execute("CREATE INDEX IF NOT EXISTS event_room ON events(room)")
    cursor.execute("CREATE INDEX IF NOT EXISTS event_track ON events(track)")
    cursor.execute("CREATE INDEX IF NOT EXISTS event_starred ON events(starred)")
    cursor.execute("""CREATE TABLE IF NOT EXISTS links (
        event_id TEXT(50),
        href TEXT,
        name TEXT,
        FOREIGN KEY(event_id) REFERENCES events(id))""")
    cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS name_event ON links(event_id, name)")


def drop(cursor):
    # Do I need this?
    ...



from gi.repository import GObject
from gi.repository import Gio


class NetworkConnectionMonitor(GObject.Object):
    __gtype_name__ = "ConfyNetworkConnectionMonitor"

    def __init__(self):
        super().__init__()
        self._isconnected = False
        self.nm = Gio.NetworkMonitor.get_default()
        self.nm.connect("network-changed", self.on_network_changed)
        self.on_network_changed()

    def on_network_changed(self, *args):
        _ic = self._isconnected
        self._isconnected = self.nm.get_connectivity() == Gio.NetworkConnectivity.FULL
        if (self._isconnected != _ic):
            # print("NetworkConnectionMonitor.isconnected", self._isconnected)
            self.notify("isconnected")

    @GObject.Property(type=bool, default=False)
    def isconnected(self):
        """The system can reach the global netowrk. Bool, read-only"""
        # print("NetworkConnectionMonitor.isconnected getter:", self._isconnected)
        return self._isconnected

    def get_isconnected(self) -> bool:
        return self._isconnected



# ics.py
#
# Copyright 2020-2022 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import html
from datetime import datetime, timedelta
from dateutil.tz import UTC
from icalendar import Calendar  # type: ignore

from .. import local
from ..models import Meta


def import_ics(content: str, url: str):
    """Import data from ICS

    all dates will be converted to UTC and stored in db without tzinfo,
    because sqlite3 driver spits errors otherwise.
    """

    cal = Calendar.from_ical(content)

    _db = local.getDb()

    meta_title = None
    meta_end = datetime(1970, 1, 1, 0, 0, 0)
    meta_start = datetime(3100, 12, 12, 23, 59, 59)

    for c in cal.walk():
        if c.name == "VCALENDAR":
            meta_title = c.get('X-WR-CALNAME')

        if c.name == "VEVENT":
            fulltextsearch = []
            eventid = c['UID']
            if c.get('DTSTART'):
                start = c['DTSTART'].dt.astimezone(UTC).replace(tzinfo=None)
            else:
                start = c['DTSTAMP'].dt.astimezone(UTC).replace(tzinfo=None)
            if c.get('DURATION'):
                end = start + c['DURATION'].dt
            elif c.get('DTEND'):
                end = c['DTEND'].dt.astimezone(UTC).replace(tzinfo=None)
            else:
                end = start + timedelta(hours=1)
            evtdate = start.date()
            room = c.get('LOCATION')
            slug = None
            title = c.get('SUMMARY', "")
            subtitle = None
            track = None
            evtype = None
            abstract = None
            description = c.get('DESCRIPTION', "")

            if title == "" and description == "":
                title = "Untitled event"
            if title == "" and description != "":
                title = description[:20]
                if len(description) > 20:
                    title = title + "…"

            if room:
                room = html.unescape(room)
            if room == "":
                room = None

            if start < meta_start:
                meta_start = start
            if end > meta_end:
                meta_end = end

            fulltextsearch += [s for s in [title, subtitle, abstract, description, room, track] if s is not None]

            links = []
            _url = c.get('URL')
            if _url:
                links.append((eventid, _url, _url))

            _db.execute("""INSERT OR REPLACE INTO events
                           (id, date, start, end, room, slug, title, subtitle, track, type, abstract, description, starred)
                           VALUES (?,?,?,?,?,?,?,?,?,?,?,?, (SELECT starred FROM events WHERE id=?))""",
                        (eventid, evtdate, start, end, room, slug, title, subtitle, track, evtype, abstract, description, eventid))
            # _db.executemany("INSERT OR REPLACE INTO persons (id, name) VALUES (?, ?)", persons)
            # _db.executemany("""INSERT OR REPLACE INTO event_person (event_id, person_id)
            #                    VALUES (?, ?) ON CONFLICT DO NOTHING""",
            #                 [ (eventid, p[0]) for p in persons ])
            _db.executemany("""INSERT OR REPLACE INTO links
                                (event_id, href, name) VALUES (?,?,?)""", links)
            fulltextsearchstr = ' '.join(fulltextsearch)
            _db.execute("""INSERT OR REPLACE INTO fts_event (event_id, text) VALUES (?, ?)""", (eventid, fulltextsearchstr))
    _db.commit()

    with Meta() as m:
        m.url = url
        m.title = meta_title
        m.last_update = time.time()  # timestamp
        m.start = meta_start.date()
        m.end = meta_end.date()



# preferences.py
#
# Copyright 2020-2022 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _

from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Adw

from .settings import Settings, CacheDuration
from . import local


def format_size(num):
    for unit in ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB']:
        if abs(num) < 1024.0:
            return "%3.1f%s" % (num, unit)
        num /= 1024.0
    return "%.1f%s" % (num, 'Yi')


class PreferencesWindow(Adw.PreferencesWindow):
    def __init__(self, **kwargs):
        opts = dict(
            default_height=290,
            default_width=640,
            search_enabled=False,
        )
        opts.update(kwargs)
        super().__init__(**opts)

        page = Adw.PreferencesPage()

        group = Adw.PreferencesGroup()

        listcache = Adw.ComboRow(
            title=_("Events list cache duration"),
            subtitle=_("Try to update list cache when it's older than this."))
        listcache.set_model(CacheDuration.as_model())
        Settings.instance().bind('list-cache', listcache, 'selected', Gio.SettingsBindFlags.DEFAULT)

        eventcache = Adw.ComboRow(
            title=_("Events info cache duration"),
            subtitle=_("Try to update event cache when it's older than this."))
        eventcache.set_model(CacheDuration.as_model())
        Settings.instance().bind('event-cache', eventcache, 'selected', Gio.SettingsBindFlags.DEFAULT)

        group.add(listcache)
        group.add(eventcache)
        page.add(group)

        group = Adw.PreferencesGroup()
        button = Gtk.Button(label=_("Clear"),
                            valign=Gtk.Align.CENTER, halign=Gtk.Align.END)
        button.get_style_context().add_class("list-button")
        row = Adw.ActionRow(
            title=_("Cached data"))
        self._update_space_label(row)
        row.add_suffix(button)
        row.set_activatable_widget(button)
        button.connect("clicked", self._delete_cached_files, row)
        group.add(row)
        page.add(group)

        self.add(page)

    def _delete_cached_files(self, btn, row):
        local.clear_cache()
        self._update_space_label(row)

    def _update_space_label(self, row):
        cache_size = local.get_cache_size()
        occupied_space = format_size(cache_size)
        row.set_subtitle(_("Confy has {} of cached data").format(occupied_space))



# widgets/datepickerrow.py
#
# Copyright 2020-2023 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import re

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Adw


class DatePickerRow(Adw.EntryRow):
    __gtype_name__ = "ConfyDatePickerRow"

    @GObject.Signal
    def date_changed(self):
        ...

    def __init__(self, **kwargs):
        self.date = None

        super().__init__(**kwargs)
        self.set_selectable(False)

        self.pop = Gtk.Popover()
        self.cal = Gtk.Calendar()
        self.cal.show()
        self.pop.set_child(self.cal)

        self.menubutton = Gtk.MenuButton(
            popover=self.pop,
            valign=Gtk.Align.CENTER,
            icon_name="x-office-calendar-symbolic"
        )
        self.menubutton.add_css_class("flat")

        self.add_suffix(self.menubutton)

        self.cal.connect("day-selected", self._cal_changed)
        self.connect("changed", self._entry_changed)
        self.set_date(None)

    def get_date(self):
        return self.date

    def get_datetime(self):
        return datetime.datetime.combine(self.date, datetime.time.min)

    def _set_date(self, date):
        _date = self.date
        self.date = date
        if self.date != _date:
            self.emit("date_changed")

        if date is None:
            lbl = ""
            gdate = GLib.DateTime.new_now_local()
        else:
            lbl = str(date)
            gdate = GLib.DateTime.new_local(date.year, date.month, date.day, 0, 0, 0)

        return lbl, gdate

    def set_date(self, date):
        lbl, gdate = self._set_date(date)
        self.cal.select_day(gdate)
        self.set_text(lbl)

    def _cal_changed(self, *args):
        date = self.cal.get_date()
        date = datetime.date(
            year=date.get_year(),
            month=date.get_month(),
            day=date.get_day_of_month()
        )
        self.set_date(date)

    def _entry_changed(self, *args):
        val = self.get_text()
        if re.match(r'\d\d\d\d-\d\d-\d\d', val):
            date = datetime.date(*[int(s) for s in val.split("-")])
            self._set_date(date)
        else:
            self._set_date(None)



# events.py
#
# Copyright 2020-2024 Fabio Comuni, et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _
from typing import Any, Optional
from dataclasses import dataclass, field

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Adw

from .details import PageDetails
from .. import models
from ..config import FMT_DAY  # type:ignore # config is generated at compile time
from ..utils import clean_markup
from ..utils import InterruptibleTimeout
from ..widgets import EventActionRow, SearchBarOverlay


@dataclass
class ListBuilderCtx:
    lastobj: Optional[models.Event] = None
    listbox: Optional[Gtk.ListBox] = None
    data: list[GObject.GObject] = field(default_factory=list)

    def __repr__(self):
        return f"<ListBuilderCtx lastobj={self.lastobj}, listbox={self.listbox}>"


@Gtk.Template(resource_path="/net/kirgroup/confy/pages/events.ui")
class PageEvents(Adw.NavigationPage):
    """Talks list page (called "Events" from pentabarf xml)"""
    __gtype_name__ = "ConfyPageEvents"

    stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore

    pagebox: Gtk.Box = Gtk.Template.Child()  # type: ignore

    # search_button: Gtk.ToggleButton = Gtk.Template.Child()  # type: ignore
    searchbar: SearchBarOverlay = Gtk.Template.Child()  # type: ignore
    scroll: Gtk.ScrolledWindow = Gtk.Template.Child()  # type: ignore

    filters: Optional[dict[str, Any]] = None
    _group_by: Optional[str] = None
    _order_by: str = "start, room"

    _update_handler_id: Optional[int] = None
    _value: Optional[models.Event] = None

    timeout = InterruptibleTimeout()

    _subtitle = ""
    _model = ""

    @GObject.Property(type=str, default="")
    def subtitle(self):
        return self._subtitle

    @subtitle.setter
    def subtitle_setter(self, v):
        self._subtitle = v

    @GObject.Property(type=str, default="")
    def model(self):
        return self._model

    @model.setter
    def model_setter(self, v):
        if v not in ["days", "tracks", "rooms", "starred", "search"]:
            raise ValueError(f"Model {v} is not valid for PageList.")
        self._model = v

    @GObject.Property(type=str, default=None)
    def group_by(self):
        return self._group_by

    @group_by.setter
    def group_by_setter(self, value: Optional[str]):
        if value == "NONE":  # I don't want to fight with Variant
            value = None
        self._group_py = value
        match value:
            case "track":
                self.order_by = "track, start, room"
            case "date":
                self.order_by = "start, room"
            case "room":
                self.order_by = "room, start"
            case _:
                self.order_by = "start, room"
        self._group_by_action.set_state(GLib.Variant.new_string(value if value else "NONE"))

    def set_model(self, v):
        self.props.model = v

    def set_group_by(self, value: Optional[str] = None):
        self.props.group_by = value  # type: ignore

    def __init__(self, value, **kwargs):
        super().__init__(**kwargs)
        self._setup_actions()
        self.set_value(value)

    def _setup_actions(self):
        group = Gio.SimpleActionGroup.new()
        # group by
        parameter_type = GLib.VariantType.new("s")
        state = GLib.Variant("s", "NONE")
        self._group_by_action = action = Gio.SimpleAction.new_stateful("group-by", parameter_type, state)
        action.connect("activate", self._on_group_by)
        group.insert(action)
        self.insert_action_group("list", group)

    def _on_group_by(self, action: Gio.SimpleAction, param: GLib.Variant):
        self.set_group_by(param.get_string())
        action.set_state(param)
        self.update()
        with models.Meta() as m:
            m.set(f"{self._model}_group_by", param.get_string())

    def get_value(self):
        return self._value

    def set_value(self, obj):
        print(f"PageEvents: set_value(obj={obj}) model={self._model}")
        if self._update_handler_id:
            models.dbUpdateWatcher.disconnect(self._update_handler_id)
        self._update_handler_id = None

        self.scroll.get_vadjustment().set_value(0)
        # self.search_button.props.visible = True
        # self.search_button.props.active = False
        self.searchbar.props.search_mode_enabled = False
        self.searchbar.props.show_close_button = True

        if self._model == "days":
            self.props.title = _("Talks on {}").format(obj.date.strftime(FMT_DAY))
            self.props.subtitle = ""
            self.filters = {'day': obj}
            self.set_group_by(None)
            self.order_by = "start, room"

        elif self._model == "tracks":
            subtitle = []
            if len(obj.room) == 1 and obj.room[0].name != "":
                subtitle.append(obj.room[0].name)
            if len(obj.date) == 1:
                subtitle.append(obj.date[0].date.strftime(FMT_DAY))

            self.props.title = _("{} Track").format(obj)
            self.props.subtitle = ", ".join(subtitle)
            self.filters = {'track': obj}
            self.set_group_by("date")

        elif self._model == "rooms":
            if obj.name is None:
                title = _("No Room")
            else:
                title = _("Room {}").format(clean_markup(obj.name))

            self.props.title = title
            self.props.subtitle = ""
            self.filters = {'room': obj}
            self.set_group_by("date")

        elif self._model == "starred":
            self.props.title = _("Starred")
            self.props.subtitle = ""
            self.filters = {'starred': True}
            self.set_group_by("date")  # TODO: or None?
            self._update_handler_id = models.dbUpdateWatcher.connect("update", self.update)

        elif self.model == "search":
            self.props.title = _("Search Talks")
            self.props.subtitle = ""
            # self.search_button.props.active = True
            # self.search_button.props.visible = False
            self.searchbar.props.search_mode_enabled = True
            self.searchbar.props.show_close_button = False
            self.filters = None
            self.set_group_by(None)

        else:
            print(f"PageEvents: set_value: no valid model set '{self._model}'")
            return

        _meta = models.Meta()
        _key = f"{self._model}_group_by"
        if _key in _meta:
            self.set_group_by(_meta.get(_key))

        self.update()

    def update(self, *args):
        count, data = self.get_objects()

        self.timeout.stop()
        # empty box
        w = self.pagebox.get_last_child()
        while w:
            self.pagebox.remove(w)
            w = self.pagebox.get_last_child()

        if count == 0:
            self.show_placeholder()
        else:
            self.stack.set_visible_child_name('list')

        ctx = ListBuilderCtx()
        ctx.data = data
        self.timeout = InterruptibleTimeout(500, self.build_list, ctx)

    def build_list(self, timeout, ctx):
        for idx, obj in enumerate(ctx.data):
            if timeout.must_stop():
                return False

            header = self.build_header(obj, ctx.lastobj)
            if header:
                ctx.listbox = None
                self.pagebox.append(header)

            row = self.build_row(obj)
            if ctx.listbox is None:
                ctx.listbox = self.build_listbox()
                self.pagebox.append(ctx.listbox)
            ctx.listbox.append(row)
            ctx.lastobj = obj
            if idx > 10:
                return True

        return False

    def build_listbox(self):
        lb = Gtk.ListBox(margin_bottom=20)
        lb.add_css_class("boxed-list")
        lb.connect("row-activated", self.on_litstbox_activated)
        return lb

    def build_header(self, current, before):
        """
        `current` is current data to be added, `before` is data
        added before.
        return a widget for a new header, or None.
        """
        _title = None
        match self._group_py:
            case "date":
                if before is None or current.date.strftime(FMT_DAY) != before.date.strftime(FMT_DAY):
                    _title = current.date.strftime(FMT_DAY)
            case "track":
                if before is None or current.track != before.track:
                    _title = current.track
                if _title == "":
                    _title = "<i>no track</i>"
            case "room":
                if before is None or current.room != before.room:
                    _title = current.room

        if _title:
            label = Gtk.Label(label=_title, halign=Gtk.Align.START, use_markup=True)
            label.add_css_class("heading")
            return label

        return None

    @Gtk.Template.Callback()
    def on_searchbar_entry_changed(self, e):
        self.update()

    @Gtk.Template.Callback()
    def on_search_mode_changed(self, *_):
        self.update()

    def get_objects(self):
        if self.model == "search":
            term = self.searchbar.entry.get_text()
            is_starred = self.searchbar.is_starred.get_active()
            _filters = {}
            if is_starred:
                _filters['starred'] = True

            g = models.Event.search(term, **_filters)
            return next(g), g
        if self.filters is None:
            return 0, []
        if self.searchbar.is_searching():
            term = self.searchbar.entry.get_text()
            is_starred = self.searchbar.is_starred.get_active()
            _filters = self.filters.copy()
            if is_starred:
                _filters['starred'] = True
            g = models.Event.search(term, order_by=self._group_py, **_filters)
            return next(g), g
        g = models.Event.filter(**self.filters, order_by=self.order_by)
        return next(g), g

    def build_row(self, obj):
        row = EventActionRow(obj)
        return row

    def show_placeholder(self):
        if self._model == "search" or self.searchbar.is_searching():
            self.stack.set_visible_child_name('no_match')
        elif self._model == "starred":
            self.stack.set_visible_child_name('empty_star')
        else:
            self.stack.set_visible_child_name('list')

    def on_litstbox_activated(self, listbox, row):
        self._value = value = row.obj
        page = PageDetails(obj=value)
        self.get_parent().push(page)
